"""
Extract code cells from notebook and save them to temporary Python file.

Markdown cells, output, and metadata are ignored.
"""

import ast
import json
import secrets
from collections import defaultdict
from typing import (
    DefaultDict,
    Iterator,
    List,
    Mapping,
    MutableMapping,
    NamedTuple,
    Optional,
    Sequence,
    Set,
    Tuple,
)

import tokenize_rt
from IPython.core.inputtransformer2 import TransformerManager

from nbqa.handle_magics import IPythonMagicType, MagicHandler, NewMagicHandler
from nbqa.notebook_info import NotebookInfo

CODE_SEPARATOR = f"# %%NBQA-CELL-SEP{secrets.token_hex(3)}\n"
MAGIC = ["time", "timeit", "capture", "pypy", "python", "python3"]
NEWLINE = "\n"
NEWLINES = defaultdict(lambda: NEWLINE * 3)
NEWLINES["isort"] = NEWLINE * 2


class Index(NamedTuple):
    """Keep track of line and cell number while iterating over cells."""

    line_number: int
    cell_number: int


def _is_ipython_magic(node: ast.expr) -> bool:
    """Check if attribute is IPython magic."""
    return (
        isinstance(node, ast.Attribute)
        and isinstance(node.value, ast.Call)
        and isinstance(node.value.func, ast.Name)
        and node.value.func.id == "get_ipython"
    )


class SystemAssignsFinder(ast.NodeVisitor):
    """Find assignments of system commands."""

    def __init__(self) -> None:
        """Record where system assigns occur."""
        self.system_assigns: Set[Tuple[int, int]] = set()

    def visit_Assign(self, node: ast.Assign) -> None:  # pylint: disable=C0103
        """Find assignments of ipython magic."""
        if isinstance(node.value, ast.Call) and _is_ipython_magic(node.value.func):
            self.system_assigns.add((node.value.lineno, node.value.col_offset))

        self.generic_visit(node)


class Visitor(ast.NodeVisitor):
    """Visit cell to look for get_ipython calls."""

    def __init__(self, system_assigns: Set[Tuple[int, int]]) -> None:
        """Magics will record where magics occur."""
        self.magics: MutableMapping[
            int, List[Tuple[int, Optional[str], Optional[str]]]
        ] = defaultdict(list)
        self.system_assigns = system_assigns

    def visit_Call(self, node: ast.Call) -> None:  # pylint: disable=C0103,R0912
        """
        Get source to replace ipython magic with.

        Parameters
        ----------
        node
            Function call.

        Raises
        ------
        AssertionError
            Defensive check.
        """
        if _is_ipython_magic(node.func):
            assert isinstance(node.func, ast.Attribute)  # help mypy
            args = []
            for arg in node.args:
                if isinstance(arg, ast.Str):
                    args.append(arg.s)
                else:
                    raise AssertionError(
                        "Please report a bug at https://github.com/nbQA-dev/nbQA/issues"
                    )
            assert (
                args
            ), "Please report a bug at https://github.com/nbQA-dev/nbQA/issues"
            magic_type: Optional[str] = None
            if node.func.attr == "run_cell_magic":
                src: Optional[str] = f"%%{args[0]}"
                if args[1]:
                    assert src is not None
                    src += f" {args[1]}"
                magic_type = "cell"
            elif node.func.attr == "run_line_magic":
                if args[0] == "pinfo":
                    src = f"{args[1]}?"
                elif args[0] == "pinfo2":
                    src = f"{args[1]}??"
                else:
                    src = f"%{args[0]}"
                    if args[1]:
                        assert src is not None
                        src += f" {args[1]}"
                magic_type = "line"
            elif node.func.attr == "system":
                src = f"!{args[0]}"
                magic_type = "line"
            elif node.func.attr == "getoutput":
                if (node.lineno, node.col_offset) in self.system_assigns:
                    src = f"!{args[0]}"
                else:
                    src = f"!!{args[0]}"
                magic_type = "line"
            else:
                src = None
            self.magics[node.lineno].append(
                (
                    node.col_offset,
                    src,
                    magic_type,
                )
            )
        self.generic_visit(node)


def _process_source(
    source: str,
    command: str,
    magic_substitutions: List[NewMagicHandler],
    *,
    skip_bad_cells: bool,
) -> str:
    """Temporarily replace ipython magics - don't process if can't."""
    try:
        ast.parse(source)
    except SyntaxError:
        pass
    else:
        # Source has no IPython magic, return it directly
        return source
    body = TransformerManager().transform_cell(source)
    if len(body.splitlines()) != len(source.splitlines()):
        handler = NewMagicHandler(source, command, magic_type=None)
        magic_substitutions.append(handler)
        return handler.replacement
    try:
        tree = ast.parse(body)
    except SyntaxError:
        if skip_bad_cells:
            handler = NewMagicHandler(source, command, magic_type=None)
            magic_substitutions.append(handler)
            return handler.replacement
        return source
    system_assigns_finder = SystemAssignsFinder()
    system_assigns_finder.visit(tree)
    visitor = Visitor(system_assigns_finder.system_assigns)
    visitor.visit(tree)
    new_src = []
    for i, line in enumerate(body.splitlines(), start=1):
        if i in visitor.magics:
            col_offset, src, magic_type = visitor.magics[i][0]
            if src is None or len(visitor.magics[i]) > 1:
                # unusual case - skip cell completely for now
                handler = NewMagicHandler(source, command, magic_type=magic_type)
                magic_substitutions.append(handler)
                return handler.replacement
            handler = NewMagicHandler(
                src,
                command,
                magic_type=magic_type,
            )
            magic_substitutions.append(handler)
            line = line[:col_offset] + handler.replacement
        new_src.append(line)
    return "\n" * (len(source) - len(source.lstrip("\n"))) + "\n".join(new_src)


def _replace_magics(
    source: Sequence[str],
    magic_substitutions: List[NewMagicHandler],
    command: str,
    *,
    skip_bad_cells: bool,
) -> Iterator[str]:
    """
    Replace IPython line magics with valid python code.

    Parameters
    ----------
    source
        Source from notebook cell.
    magic_substitutions
        List to store all the ipython magics substitutions

    Yields
    ------
    str
        Line from cell, with line magics replaced with python code
    """
    # if first line is cell magic, process it separately
    if MagicHandler.get_ipython_magic_type(source[0]) == IPythonMagicType.CELL:
        header = _process_source(
            source[0], command, magic_substitutions, skip_bad_cells=skip_bad_cells
        )
        cell = _process_source(
            "".join(source[1:]),
            command,
            magic_substitutions,
            skip_bad_cells=skip_bad_cells,
        )
        yield "\n".join([header, cell])
    else:
        yield _process_source(
            "".join(source), command, magic_substitutions, skip_bad_cells=skip_bad_cells
        )


def _parse_cell(
    source: Sequence[str],
    cell_number: int,
    temporary_lines: MutableMapping[int, Sequence[NewMagicHandler]],
    command: str,
    *,
    skip_bad_cells: bool,
) -> str:
    """
    Parse cell, replacing line magics with python code as placeholder.

    Parameters
    ----------
    source
        Source from notebook cell.
    cell_number
        Number identifying the notebook cell.
    temporary_lines
        Mapping to store the cell number to all the ipython magics replaced in those cells.
    command
        The third-party tool being run.

    Returns
    -------
    str
        Parsed cell.
    """
    substituted_magics: List[NewMagicHandler] = []
    parsed_cell = CODE_SEPARATOR

    for parsed_line in _replace_magics(
        source, substituted_magics, command, skip_bad_cells=skip_bad_cells
    ):
        parsed_cell += parsed_line

    if substituted_magics:
        temporary_lines[cell_number] = substituted_magics

    return f"{parsed_cell}{NEWLINES[command]}"


def _get_line_numbers_for_mapping(
    cell_source: str, magic_substitutions: Sequence[NewMagicHandler]
) -> Mapping[int, int]:
    """Get the line number mapping from python file to notebook cell.

    Parameters
    ----------
    cell_source
        Source code of the notebook cell
    magic_substitutions
        IPython magics substituted in the current notebook cell.

    Returns
    -------
    Mapping[int, int]
        Line number mapping from temporary python file to notebook cell
    """
    lines_in_cell = cell_source.splitlines()
    line_mapping: MutableMapping[int, int] = {}

    if not magic_substitutions:
        line_mapping.update({i: i for i in range(len(lines_in_cell))})
    else:
        line_number = -1

        for line_no, _ in enumerate(lines_in_cell):
            line_number += 1
            line_mapping[line_no] = line_number

    return line_mapping


def _should_ignore_code_cell(
    source: Sequence[str], process_cells: Sequence[str]
) -> bool:
    """
    Return True if the current cell should be ignored from processing.

    Parameters
    ----------
    source
        Source from the notebook cell
    process_cells
        Extra cells which nbqa should process.

    Returns
    -------
    bool
        True if the cell should ignored else False
    """
    if not source:
        return True
    process = MAGIC + [i.strip() for i in process_cells]
    magic_type = MagicHandler.get_ipython_magic_type(source[0])
    if magic_type != IPythonMagicType.CELL:
        return False
    first_line = source[0].lstrip()
    return first_line.split()[0] not in {f"%%{magic}" for magic in process}


def _has_trailing_semicolon(src: str) -> Tuple[str, bool]:
    """
    Check if cell has trailing semicolon.

    Parameters
    ----------
    src
        Notebook cell source.

    Returns
    -------
    bool
        Whether notebook has trailing semicolon.
    """
    tokens = tokenize_rt.src_to_tokens(src)
    trailing_semicolon = False
    for idx, token in tokenize_rt.reversed_enumerate(tokens):
        if not token.src.strip(" \n") or token.name == "COMMENT":
            continue
        if token.name == "OP" and token.src == ";":
            tokens[idx] = token._replace(src="")
            trailing_semicolon = True
        break
    if not trailing_semicolon:
        return src, False
    return tokenize_rt.tokens_to_src(tokens), True


def main(  # pylint: disable=R0914
    notebook: str,
    file_descriptor: int,
    process_cells: Sequence[str],
    command: str,
    *,
    skip_bad_cells: bool,
) -> NotebookInfo:
    """
    Extract code cells from notebook and save them in temporary Python file.

    Parameters
    ----------
    notebook
        Jupyter Notebook third-party tool is being run against.
    process_cells
        Extra cells which nbqa should process.
    command
        The third-party tool being run.

    Returns
    -------
    NotebookInfo

    """
    with open(str(notebook), encoding="utf-8") as handle:
        content = handle.read()

    cells = json.loads(content)["cells"]

    result = []
    cell_mapping = {0: "cell_0:0"}
    index = Index(line_number=0, cell_number=0)
    trailing_semicolons = set()
    temporary_lines: DefaultDict[int, Sequence[NewMagicHandler]] = defaultdict(list)
    code_cells_to_ignore = set()

    for cell in cells:
        if cell["cell_type"] == "code":
            index = index._replace(cell_number=index.cell_number + 1)

            if _should_ignore_code_cell(cell["source"], process_cells):
                code_cells_to_ignore.add(index.cell_number)
                continue

            parsed_cell = _parse_cell(
                cell["source"],
                index.cell_number,
                temporary_lines,
                command,
                skip_bad_cells=skip_bad_cells,
            )

            cell_mapping.update(
                {
                    py_line
                    + index.line_number
                    + 1: f"cell_{index.cell_number}:{cell_line}"
                    for py_line, cell_line in _get_line_numbers_for_mapping(
                        parsed_cell, temporary_lines[index.cell_number]
                    ).items()
                }
            )
            parsed_cell, trailing_semicolon = _has_trailing_semicolon(parsed_cell)
            if trailing_semicolon:
                trailing_semicolons.add(index.cell_number)
            result.append(parsed_cell)
            index = index._replace(
                line_number=index.line_number + len(parsed_cell.splitlines())
            )

    result_txt = "".join(result).rstrip(NEWLINE) + NEWLINE if result else ""
    with open(file_descriptor, "w", encoding="utf-8") as handle:
        handle.write(result_txt)

    return NotebookInfo(
        cell_mapping, trailing_semicolons, temporary_lines, code_cells_to_ignore
    )



"""Detect ipython magics and provide python code replacements for those magics."""
import secrets
from abc import ABC
from enum import Enum
from typing import ClassVar, Mapping, Optional, Sequence

from IPython.core.inputtransformer2 import TransformerManager

COMMANDS_WITH_STRING_TOKEN = {"flake8"}


class IPythonMagicType(Enum):
    """Enumeration representing various types of IPython magics."""

    SHELL = 0
    HELP = 1
    LINE = 2
    CELL = 3
    NO_MAGIC = 4


class NewMagicHandler:  # pylint: disable=R0903
    """Handle different types of magics."""

    def __init__(self, src: str, command: str, magic_type: Optional[str]):
        """
        Handle magic.

        Parameters
        ----------
        src
            Original code
        command
            e.g. flake8
        magic_type
            E.g. cell, line, ...
        """
        self.src = src
        token = secrets.token_hex(4)
        if command in COMMANDS_WITH_STRING_TOKEN:
            self.token = f'"{token}"'
        else:
            self.token = f"0x{int(token, base=16):X}"
        if magic_type == "cell":
            self.replacement = f"# CELL MAGIC {self.token}"
        else:
            self.replacement = f"hash({self.token})"


class MagicHandler(ABC):  # pylint: disable=R0903
    """Base class of different types of magic handlers."""

    # To Developers: Its better to preserve the order in which
    # this dictionary is populated. For instance, if HELP is inserted
    # after LINE leads to bug, since they both share the same prefix.
    _MAGIC_PREFIXES: ClassVar[Mapping[IPythonMagicType, Sequence[str]]] = {
        IPythonMagicType.SHELL: ["get_ipython().system", "get_ipython().getoutput"],
        IPythonMagicType.HELP: ["get_ipython().run_line_magic('pinfo"],
        IPythonMagicType.LINE: ["get_ipython().run_line_magic", "get_ipython().magic"],
        IPythonMagicType.CELL: ["get_ipython().run_cell_magic"],
    }

    @staticmethod
    def get_ipython_magic_type(ipython_magic: str) -> Optional[IPythonMagicType]:
        """Return the type of ipython magic.

        This function assumes the input parameter to be a ipython magic. It is
        recommended to call this method after checking `is_ipython_magic`.

        Parameters
        ----------
        ipython_magic : str
            Ipython magic statement

        Returns
        -------
        Optional[IPythonMagicType]
            Type of the IPython magic
        """
        python_code = TransformerManager().transform_cell(ipython_magic)
        magic_type: Optional[IPythonMagicType] = None
        for magic, prefixes in MagicHandler._MAGIC_PREFIXES.items():
            if any(prefix in python_code for prefix in prefixes):
                magic_type = magic
                break
        else:
            magic_type = IPythonMagicType.NO_MAGIC

        return magic_type



"""Parses the command line arguments provided."""
import argparse
from textwrap import dedent
from typing import Optional, Sequence

from nbqa import __version__
from nbqa.text import BOLD, RESET

CONFIGURATION_URL = "https://nbqa.readthedocs.io/en/latest/configuration.html"
DOCS_URL = "https://nbqa.readthedocs.io/en/latest/index.html"
USAGE_MSG = dedent(
    f"""\
    nbqa <code quality tool> <notebook or directory> <nbqa options> <code quality tool arguments>

    {BOLD}Please specify:{RESET}
    - 1) a code quality tool (e.g. `black`, `pyupgrade`, `flake`, ...)
    - 2) some notebooks (or, if supported by the tool, directories)
    - 3) (optional) flags for nbqa (e.g. `--nbqa-mutate`)
    - 4) (optional) flags for code quality tool (e.g. `--line-length` for `black`)

    {BOLD}Examples:{RESET}
        nbqa flake8 notebook.ipynb
        nbqa black notebook.ipynb --line-length=96
        nbqa pyupgrade notebook_1.ipynb notebook_2.ipynb

    {BOLD}Mutation:{RESET} to let `nbqa` modify your notebook(s), also pass `--nbqa-mutate`, e.g.:
        nbqa black notebook.ipynb --nbqa-mutate

    See {DOCS_URL} for more details on how to run `nbqa`.
    """
)


class CLIArgs:  # pylint: disable=R0902,R0903
    """
    Stores the command line arguments passed.

    Attributes
    ----------
    command
        The third-party tool to run (e.g. :code:`mypy`).
    root_dirs
        The notebooks or directories to run third-party tool on.
    nbqa_mutate
        Whether to allow 3rd party tools to modify notebooks.
    nbqa_process_cells
        Process code within these cell magics.
    nbqa_addopts
        Any additional flags passed to third-party tool (e.g. :code:`--quiet`).
    nbqa_files
        Global file include pattern.
    nbqa_exclude
        Global file exclude pattern.
    nbqa_skip_bad_cells
        Skip cells with syntax errors.
    """

    command: str
    root_dirs: Sequence[str]
    nbqa_addopts: Sequence[str]
    nbqa_mutate: bool
    nbqa_process_cells: Optional[str]
    nbqa_diff: bool
    nbqa_files: Optional[str]
    nbqa_exclude: Optional[str]
    nbqa_skip_bad_cells: bool

    def __init__(self, args: argparse.Namespace, cmd_args: Sequence[str]) -> None:
        """
        Initialize this instance with the parsed command line arguments.

        Parameters
        ----------
        args
            Command line arguments passed to nbqa
        cmd_args
            Additional options to pass to the tool
        """
        self.command = args.command
        self.root_dirs = args.root_dirs
        self.nbqa_addopts = cmd_args
        self.nbqa_mutate = args.nbqa_mutate or False
        self.nbqa_process_cells = args.nbqa_process_cells or None
        self.nbqa_diff = args.nbqa_diff or False
        self.nbqa_files = args.nbqa_files or None
        self.nbqa_exclude = args.nbqa_exclude or None
        self.nbqa_skip_bad_cells = args.nbqa_skip_bad_cells or False

    @staticmethod
    def parse_args(argv: Optional[Sequence[str]]) -> "CLIArgs":
        """
        Parse command-line arguments.

        Parameters
        ----------
        argv
            Passed via command-line.
        Returns
        -------
        CLIArgs
            Object that holds all the parsed command line arguments.
        """
        parser = argparse.ArgumentParser(
            description="Run any standard Python code-quality tool on a Jupyter notebook.",
            usage=USAGE_MSG,
        )
        parser.add_argument("command", help="Command to run, e.g. `flake8`.")
        parser.add_argument(
            "root_dirs", nargs="+", help="Notebooks or directories to run command on."
        )
        parser.add_argument(
            "--nbqa-files",
            help="Global file include pattern.",
        )
        parser.add_argument(
            "--nbqa-exclude",
            help="Global file exclude pattern.",
        )
        parser.add_argument(
            "--nbqa-mutate",
            action="store_true",
            help="Allows `nbqa` to modify notebooks.",
        )
        parser.add_argument(
            "--nbqa-diff",
            action="store_true",
            help="Show diff which would result from running --nbqa-mutate.",
        )
        parser.add_argument(
            "--nbqa-process-cells",
            required=False,
            help=dedent(
                r"""
                Process code within these cell magics. You can pass multiple options,
                e.g. `nbqa black my_notebook.ipynb --nbqa-process-cells add_to,write_to`
                by placing commas between them.
                """
            ),
        )
        parser.add_argument(
            "--version", action="version", version=f"nbqa {__version__}"
        )
        parser.add_argument(
            "--nbqa-skip-bad-cells",
            action="store_true",
            help="Skip cells with invalid syntax.",
        )
        args, cmd_args = parser.parse_known_args(argv)
        return CLIArgs(args, cmd_args)



"""Store information about the code cells for processing."""

from typing import Mapping, Sequence, Set

from nbqa.handle_magics import NewMagicHandler


class NotebookInfo:
    """
    Store information about notebook cells used for processing.

    Attributes
    ----------
    cell_mapping
        Mapping from Python line numbers to Jupyter notebooks cells.
    trailing_semicolons
        Cell numbers where there were originally trailing semicolons.
    temporary_lines
        Mapping from cell number to all the magics substituted in those cell.
    code_cells_to_ignore
        List of code cell to ignore when modifying the source notebook.
    """

    _cell_mappings: Mapping[int, str] = {}
    _trailing_semicolons: Set[int] = set()
    _temporary_lines: Mapping[int, Sequence[NewMagicHandler]] = {}
    _code_cells_to_ignore: Set[int] = set()

    def __init__(
        self,
        cell_mappings: Mapping[int, str],
        trailing_semicolons: Set[int],
        temporary_lines: Mapping[int, Sequence[NewMagicHandler]],
        code_cells_to_ignore: Set[int],
    ) -> None:
        """
        Initialize current NotebookInfo instance.

        Parameters
        ----------
        cell_mappings
            Mapping from Python line numbers to Jupyter notebooks cells.
        trailing_semicolons
            Cell numbers where there were originally trailing semicolons.
        temporary_lines
            Mapping from cell number to all the magics substituted in those cell.
        code_cells_to_ignore
            List of cell numbers to ignore when modifying the source notebook.
        """
        self._cell_mappings = cell_mappings
        self._trailing_semicolons = trailing_semicolons
        self._temporary_lines = temporary_lines
        self._code_cells_to_ignore = code_cells_to_ignore

    @property
    def cell_mappings(self) -> Mapping[int, str]:
        """Return mapping from Python line numbers to Jupyter notebooks cells."""
        return self._cell_mappings

    @property
    def trailing_semicolons(self) -> Set[int]:
        """Return cell numbers where there were originally trailing semicolons."""
        return self._trailing_semicolons

    @property
    def temporary_lines(self) -> Mapping[int, Sequence[NewMagicHandler]]:
        """Return mapping from cell number to all the magics substituted."""
        return self._temporary_lines

    @property
    def code_cells_to_ignore(self) -> Set[int]:
        """Return code cell numbers to ignore when modifying the source notebook."""
        return self._code_cells_to_ignore



"""
Find project root.

Taken from https://github.com/psf/black/blob/master/src/black/__init__.py
"""

from functools import lru_cache
from pathlib import Path
from typing import Iterable

# files and folders known to indicate a project root
KNOWN_PROJECT_ROOT_DIRS = [".git", ".hg"]
KNOW_PROJECT_ROOT_FILES = [
    ".nbqa.ini",
    "setup.py",
    "setup.cfg",
    "pyproject.toml",
    "MANIFEST.in",
]


@lru_cache()
def find_project_root(
    srcs: Iterable[str],
    root_files: Iterable[str] = tuple(KNOW_PROJECT_ROOT_FILES),
    root_dirs: Iterable[str] = tuple(KNOWN_PROJECT_ROOT_DIRS),
) -> Path:
    """
    Return a directory containing .git, .hg, or nbqa.ini.

    That directory will be a common parent of all files and directories
    passed in `srcs`.
    If no directory in the tree contains a marker that would specify it's the
    project root, the root of the file system is returned.

    Parameters
    ----------
    srcs
        Source paths.
    root_files
        Files indicating that the current directory is the project root.

    Returns
    -------
    Path
        Project root.
    """
    path_srcs = [Path(Path.cwd(), src).resolve() for src in srcs]

    # A list of lists of parents for each 'src'. 'src' is included as a
    # "parent" of itself if it is a directory
    src_parents = [
        list(path.parents) + ([path] if path.is_dir() else []) for path in path_srcs
    ]

    common_base = max(
        set.intersection(*(set(parents) for parents in src_parents)),
        key=lambda path: path.parts,
    )

    for directory in (common_base, *common_base.parents):

        for known_project_root_dir in root_dirs:
            if (directory / known_project_root_dir).is_dir():
                return directory
        for know_project_root_file in root_files:
            if (directory / know_project_root_file).is_file():
                return directory

    return Path("/").resolve()



"""Parse nbqa configuration from TOML file."""
from pathlib import Path
from typing import Any, Mapping, Optional

import toml

from nbqa.config.config import Configs

_ROOT_CONFIG_KEY: str = "tool"
_NBQA_CONFIG_KEY: str = "nbqa"


def parse_from_pyproject_toml(command: str, file_path: Path) -> Optional[Configs]:
    """
    Parse nbqa configuration from TOML file.

    Parameters
    ----------
    command : str
        Third party tool to run
    file_path : Path
        Configuration file path

    Returns
    -------
    Optional[Configs]
        Config object parsed from the configuration file(if present)
    """
    config: Optional[Configs] = None
    toml_config: Mapping[str, Any] = toml.load(str(file_path))
    nbqa_toml_config: Optional[Mapping[str, Mapping[str, Any]]] = toml_config.get(
        _ROOT_CONFIG_KEY, {}
    ).get(_NBQA_CONFIG_KEY, None)

    if nbqa_toml_config is not None:
        config = Configs()

        for section in Configs.CONFIG_SECTION_PARSERS:
            if section in nbqa_toml_config:
                config.set_config(section, nbqa_toml_config[section].get(command, None))

    return config



"""Default configurations."""
from typing import Iterable, Mapping

from nbqa.save_source import CODE_SEPARATOR

DEFAULT_CONFIG: Mapping[str, Mapping[str, Iterable[str]]] = {
    "addopts": {"isort": ("--treat-comment-as-code", CODE_SEPARATOR.rstrip("\n"))},
    "config": {},
    "mutate": {},
    "process_cells": {},
    "diff": {},
    "files": {},
    "exclude": {},
    "skip_bad_cells": {},
}



"""Utility functions to deal with paths."""
from pathlib import Path
from typing import Tuple


def remove_suffix(string: str, suffix: str) -> str:
    """
    Remove suffix from string.

    Parameters
    ----------
    string
        Given string to remove suffix from.
    suffix
        Suffix to remove.

    Raises
    ------
    AssertionError
        If string doesn't end with given suffix.
    """
    if string.endswith(suffix):
        string = string[: -len(suffix)]
    else:  # pragma: nocover
        raise AssertionError(f"{string} doesn't end with suffix {suffix}")
    return string


def get_relative_and_absolute_paths(path: str) -> Tuple[str, str]:
    """Get relative (if possible) and absolute versions of path."""
    absolute_path = Path(path).resolve()
    try:
        relative_path = absolute_path.relative_to(Path.cwd())
    except ValueError:
        relative_path = absolute_path
    return str(relative_path), str(absolute_path)



"""Parse nbqa configuration from one of the supported configuration files."""

from configparser import ConfigParser
from pathlib import Path
from typing import Callable, Optional, Sequence, Tuple

from nbqa.cmdline import CLIArgs
from nbqa.config.config import Configs
from nbqa.toml_parser import parse_from_pyproject_toml


def _parse_nbqa_ini_config(command: str, file_path: Path) -> Optional[Configs]:
    """
    Parse configuration from .nbqa.ini file.

    Parameters
    ----------
    command : str
        Third party tool to run
    file_path : Path
        Configuration file path

    Returns
    -------
    Optional[Configs]
        Config object parsed from the configuration file(if present)
    """
    config_parser = ConfigParser()
    config_parser.read(file_path)

    config: Configs = Configs()
    section: str = command

    for option in config_parser.options(section):
        config.set_config(option, config_parser.get(section, option, fallback=None))

    return config


def _parse_setupcfg_or_toxini_config(
    command: str, file_path: Path
) -> Optional[Configs]:
    """
    Parse nbqa configuration from setup.cfg or tox.ini.

    Parameters
    ----------
    command : str
        Third party tool to run
    file_path : Path
        Configuration file path

    Returns
    -------
    Optional[Configs]
        Config object parsed from the configuration file(if present)
    """
    config_parser = ConfigParser()
    config_parser.read(file_path)

    config: Optional[Configs] = None
    option: str = command

    # Check if this file contains the nbqa configuration
    for section in Configs.CONFIG_SECTION_PARSERS:
        section_name = f"{CONFIG_PREFIX}{section}"
        if config_parser.has_section(section_name):
            config = Configs()
            break

    if config is not None:
        for section in Configs.CONFIG_SECTION_PARSERS:
            section_name = f"{CONFIG_PREFIX}{section}"

            # We might not find the setting for a particular tool in any of the sections
            if config_parser.has_option(section_name, option):
                config.set_config(
                    section, config_parser.get(section_name, option, fallback=None)
                )

    return config


_ConfigHandler = Callable[..., Optional[Configs]]

CONFIG_PREFIX: str = "nbqa."

_CONFIG_FILE_HANDLERS: Sequence[Tuple[str, Optional[_ConfigHandler]]] = [
    ("pyproject.toml", parse_from_pyproject_toml),
    ("setup.cfg", _parse_setupcfg_or_toxini_config),
    ("tox.ini", _parse_setupcfg_or_toxini_config),
    (".nbqa.ini", _parse_nbqa_ini_config),
]


def parse_config_from_file(cli_args: CLIArgs, project_root: Path) -> Optional[Configs]:
    """
    Find the file that contains nbqa configuration and parse the configuration from it.

    Parameters
    ----------
    cli_args : CLIArgs
        Commandline arguments passed to nbqa
    project_root : Path
        Root of the repository or project

    Returns
    -------
    Optional[Configs]
        Configuration read from the file(if any)
    """
    config: Optional[Configs] = None

    for config_file, config_handler in _CONFIG_FILE_HANDLERS:
        file_path: Path = project_root / config_file

        if file_path.is_file() and config_handler is not None:
            config = config_handler(cli_args.command, file_path)

        # we found the config. skip other files
        if config is not None:
            break

    return config



"""Module responsible for storing and handling nbqa configuration."""

from shlex import split
from textwrap import dedent
from typing import Any, Callable, ClassVar, Dict, Optional, Sequence, Union

from nbqa.cmdline import CLIArgs
from nbqa.config.default_config import DEFAULT_CONFIG

ConfigParser = Callable[[str], Union[str, bool, Sequence[str]]]


class Configs:
    """Nbqa configuration."""

    CONFIG_SECTION_PARSERS: ClassVar[Dict[str, ConfigParser]] = {}
    CONFIG_SECTION_PARSERS["addopts"] = (
        lambda arg: split(arg) if isinstance(arg, str) else arg
    )
    CONFIG_SECTION_PARSERS["process_cells"] = (
        lambda arg: arg.split(",") if isinstance(arg, str) else arg
    )
    CONFIG_SECTION_PARSERS["mutate"] = bool
    CONFIG_SECTION_PARSERS["diff"] = bool
    CONFIG_SECTION_PARSERS["files"] = str
    CONFIG_SECTION_PARSERS["exclude"] = str
    CONFIG_SECTION_PARSERS["skip_bad_cells"] = bool

    _mutate: bool = False
    _process_cells: Sequence[str] = []
    _addopts: Sequence[str] = []
    _diff: bool = False
    _files: Optional[str] = None
    _exclude: Optional[str] = None
    _skip_bad_cells: bool = False

    def set_config(self, config: str, value: Any) -> None:
        """
        Store value for the config.

        Parameters
        ----------
        config
            Config setting supported by nbqa
        value
            Config value
        """
        if value:
            self.__setattr__(f"_{config}", self.CONFIG_SECTION_PARSERS[config](value))

    @property
    def nbqa_mutate(self) -> bool:
        """Flag if nbqa is allowed to modify the original notebook(s)."""
        return self._mutate

    @property
    def nbqa_diff(self) -> bool:
        """Show diff instead of replacing notebook code."""
        return self._diff

    @property
    def nbqa_addopts(self) -> Sequence[str]:
        """Additional options to be passed to the third party command to run."""
        return self._addopts

    @property
    def nbqa_process_cells(self) -> Sequence[str]:
        """Additional cells which nbqa should process."""
        return self._process_cells

    @property
    def nbqa_files(self) -> Optional[str]:
        """Additional cells which nbqa should ignore."""
        return self._files

    @property
    def nbqa_exclude(self) -> Optional[str]:
        """Additional cells which nbqa should ignore."""
        return self._exclude

    @property
    def nbqa_skip_cells(self) -> bool:
        """Skip cells with syntax errors."""
        return self._skip_bad_cells

    def merge(self, other: "Configs") -> "Configs":
        """
        Merge another Config instance with this instance.

        Returns
        -------
        Configs
            Merged configuration object
        """
        config: Configs = Configs()

        config.set_config("addopts", [*self._addopts, *other.nbqa_addopts])
        config.set_config(
            "process_cells",
            self._process_cells or other.nbqa_process_cells,
        )
        config.set_config("mutate", self._mutate or other.nbqa_mutate)
        config.set_config("diff", self._diff or other.nbqa_diff)
        config.set_config("files", self._files or other.nbqa_files)
        config.set_config("exclude", self._exclude or other.nbqa_exclude)
        config.set_config(
            "skip_bad_cells",
            self._skip_bad_cells or other.nbqa_skip_cells,
        )
        return config

    @staticmethod
    def parse_from_cli_args(cli_args: CLIArgs) -> "Configs":
        """
        Parse nbqa configuration from command line arguments.

        Returns
        -------
        Configs
            Configuration passed via command line arguments.
        """
        config: Configs = Configs()

        config.set_config("addopts", cli_args.nbqa_addopts)
        config.set_config("process_cells", cli_args.nbqa_process_cells)
        config.set_config("mutate", cli_args.nbqa_mutate)
        config.set_config("diff", cli_args.nbqa_diff)
        config.set_config("files", cli_args.nbqa_files)
        config.set_config("exclude", cli_args.nbqa_exclude)
        config.set_config("skip_bad_cells", cli_args.nbqa_skip_bad_cells)

        return config

    @staticmethod
    def get_default_config(command: str) -> "Configs":
        """Merge defaults."""
        defaults: Configs = Configs()
        defaults.set_config("addopts", DEFAULT_CONFIG["addopts"].get(command))
        defaults.set_config(
            "process_cells", DEFAULT_CONFIG["process_cells"].get(command)
        )
        defaults.set_config("mutate", DEFAULT_CONFIG["mutate"].get(command))
        defaults.set_config("diff", DEFAULT_CONFIG["diff"].get(command))
        defaults.set_config("files", DEFAULT_CONFIG["files"].get(command))
        defaults.set_config("exclude", DEFAULT_CONFIG["exclude"].get(command))
        defaults.set_config(
            "skip_bad_cells",
            DEFAULT_CONFIG["skip_bad_cells"].get(command),
        )

        return defaults

    def validate(self) -> None:
        """
        Check specified configs are valid.

        Raises
        ------
        ValueError
            If both --nbqa-diff and --nbqa-mutate are used together.
        """
        if self._diff and self._mutate:
            raise ValueError(
                dedent(
                    """\
                    Cannot use both `--nbqa-diff` and `--nbqa-mutate` flags together!

                    Use `--nbqa-diff` to preview changes, and `--nbqa-mutate` to apply them.
                    """
                )
            )



"""Run third-party tool (e.g. :code:`mypy`) against notebook or directory."""
import os
import re
import subprocess
import sys
import tempfile
from importlib import import_module
from pathlib import Path
from textwrap import dedent
from typing import (
    Dict,
    Iterator,
    Mapping,
    MutableMapping,
    NamedTuple,
    Optional,
    Sequence,
    Set,
    Tuple,
)

from pkg_resources import parse_version

from nbqa import config_parser, replace_source, save_source
from nbqa.cmdline import CLIArgs
from nbqa.config.config import Configs
from nbqa.find_root import find_project_root
from nbqa.notebook_info import NotebookInfo
from nbqa.optional import metadata
from nbqa.output_parser import Output, map_python_line_to_nb_lines
from nbqa.path_utils import get_relative_and_absolute_paths, remove_suffix
from nbqa.text import BOLD, RESET

BASE_ERROR_MESSAGE = dedent(
    f"""\
    {BOLD}{{}}
    Please report a bug at https://github.com/nbQA-dev/nbQA/issues {RESET}
    """
)
MIN_VERSIONS = {"isort": "5.3.0"}
VIRTUAL_ENVIRONMENTS_URL = (
    "https://realpython.com/python-virtual-environments-a-primer/"
)
EXCLUDES = (
    r"/("
    r"\.direnv|\.eggs|\.git|\.hg|\.ipynb_checkpoints|\.mypy_cache|\.nox|\.svn|\.tox|\.venv|"
    r"_build|buck-out|build|dist|venv"
    r")/"
)

REPLACE_FUNCTION = {
    True: replace_source.diff,
    False: replace_source.mutate,
}


class TemporaryFile(NamedTuple):
    """Temporary file and file descriptor."""

    fd: int
    file: str


class UnsupportedPackageVersionError(Exception):
    """Raise if installed module is older than minimum required version."""

    def __init__(self, command: str, current_version: str, min_version: str) -> None:
        """Initialise with command, current version, and minimum version."""
        self.msg = (
            f"{BOLD}nbqa only works with {command} >= {min_version}, "
            f"while you have {current_version} installed.{RESET}"
        )
        super().__init__(self.msg)


def _get_notebooks(root_dir: str) -> Iterator[Path]:
    """
    Get generator with all notebooks in directory.

    Parameters
    ----------
    root_dir
        Notebook or directory to run third-party tool on.

    Returns
    -------
    notebooks
        All Jupyter Notebooks found in directory.
    """
    if not os.path.isdir(root_dir):
        return iter((Path(root_dir),))
    return (
        i
        for i in Path(root_dir).rglob("*.ipynb")
        if not re.search(EXCLUDES, str(i.resolve().as_posix()))
    )


def _filter_by_include_exclude(
    notebooks: Iterator[Path],
    include: Optional[str],
    exclude: Optional[str],
) -> Iterator[str]:
    """
    Include files which match include, exclude those matching exclude.

    notebooks
        Notebooks (not directories) to run code quality tool on.
    include:
        Global file include pattern.
    exclude:
        Global file exclude pattern.

    Returns
    -------
    Iterator
        Notebooks matching include and not matching exclude.
    """
    include = include or ""
    exclude = exclude or "^$"
    include_re, exclude_re = re.compile(include), re.compile(exclude)
    return (
        str(notebook)
        for notebook in notebooks
        if include_re.search(str(notebook.as_posix()))
        if not exclude_re.search(str(notebook.as_posix()))
    )


def _get_all_notebooks(
    root_dirs: Sequence[str], files: Optional[str], exclude: Optional[str]
) -> Iterator[str]:
    """
    Get generator with all notebooks passed in via the command-line, applying exclusions.

    Parameters
    ----------
    root_dirs
        All the notebooks/directories passed in via the command-line.
    files
        Pattern of files to include.
    exclude
        Pattern of files to exclude.

    Returns
    -------
    Iterator
        All Jupyter Notebooks found in all passed directories/notebooks.
    """
    return _filter_by_include_exclude(
        (j for i in root_dirs for j in _get_notebooks(i)), files, exclude
    )


def _replace_temp_python_file_references_in_out_err(
    temp_python_file: str,
    notebook: str,
    out: str,
    err: str,
) -> Output:
    """
    Replace references to temporary Python file with references to notebook.

    Parameters
    ----------
    temp_python_file
        Temporary Python file where notebook was converted to.
    notebook
        Original Jupyter notebook.
    out
        Captured stdout from third-party tool.
    err
        Captured stderr from third-party tool.

    Returns
    -------
    Output
        Stdout, stderr with temporary directory replaced by current working directory.
    """
    py_basename = os.path.basename(temp_python_file)
    nb_basename = os.path.basename(notebook)
    out = out.replace(py_basename, nb_basename)
    err = err.replace(py_basename, nb_basename)

    out = out.replace(
        remove_suffix(py_basename, ".py"), remove_suffix(nb_basename, ".ipynb")
    )
    err = err.replace(
        remove_suffix(py_basename, ".py"), remove_suffix(nb_basename, ".ipynb")
    )

    return Output(out, err)


def _get_mtimes(arg: str) -> Set[float]:
    """
    Get the modification times of any converted notebooks.

    Parameters
    ----------
    arg
        Notebook to run 3rd party tool on.

    Returns
    -------
    Set
        Modification times of any converted notebooks.
    """
    return {os.path.getmtime(arg)}


def _run_command(
    command: str,
    cmd_args: Sequence[str],
    args: Sequence[str],
) -> Tuple[Output, int, bool]:
    """
    Run third-party tool against given file or directory.

    Parameters
    ----------
    command
        Third-party tool (e.g. :code:`mypy`) to run against notebook.
    cmd_args
        Flags to pass to third-party tool (e.g. :code:`--verbose`).
    args
        Notebooks, or directories of notebooks, third-party tool is being run on.

    Returns
    -------
    output
        Captured stdout, stderr from running third-party tool.
    output_code
        Return code from third-party tool.
    mutated
        Whether 3rd party tool modified any files.

    Raises
    ------
    ValueError
        If third-party tool isn't found in system.
    """
    before = [_get_mtimes(i) for i in args]

    output = subprocess.run(
        [sys.executable, "-m", command, *args, *cmd_args],
        stderr=subprocess.PIPE,
        stdout=subprocess.PIPE,
        universal_newlines=True,  # from Python3.7 this can be replaced with `text`
    )

    mutated = [_get_mtimes(i) for i in args] != before

    output_code = output.returncode

    out = output.stdout
    err = output.stderr

    return Output(out, err), output_code, mutated


def _get_command_not_found_msg(command: str) -> str:
    """Return the message to display when the command is not found by nbqa.

    Parameters
    ----------
    command : str
        Command passed to nbqa to find.

    Returns
    -------
    str
        Message to display to stdout.
    """
    template = dedent(
        f"""\
        {BOLD}Command `{command}` not found by nbqa.{RESET}

        Please make sure you have it installed in the same Python environment as nbqa. See
        e.g. {VIRTUAL_ENVIRONMENTS_URL} for how to set up
        a virtual environment in Python.

        Since nbqa is installed at {{nbqa_loc}} and uses the Python executable found at
        {{python}}, you could fix this issue by running `{{python}} -m pip install {command}`.
        """
    )
    python_executable = sys.executable
    nbqa_loc = str(Path(sys.modules["nbqa"].__file__).parent)

    return template.format(python=python_executable, nbqa_loc=nbqa_loc)


def _get_configs(cli_args: CLIArgs, project_root: Path) -> Configs:
    """
    Deal with extra configs for 3rd party tool.

    Parameters
    ----------
    cli_args
        Commandline arguments passed to nbqa
    project_root
        Root of repository, where .git / .hg / .nbqa.ini file is.

    Returns
    -------
    Configs
        Taken from CLI (if given), else from .nbqa.ini.
    """
    cli_config: Configs = Configs.parse_from_cli_args(cli_args)
    file_config: Optional[Configs] = config_parser.parse_config_from_file(
        cli_args, project_root
    )

    if file_config is not None:
        cli_config = cli_config.merge(file_config)

    return cli_config.merge(Configs.get_default_config(cli_args.command))


def _clean_up_tmp_files(nb_to_py_mapping: Mapping[str, Tuple[int, str]]) -> None:
    """Remove temporary files."""
    for file_descriptor, tmp_path in nb_to_py_mapping.values():
        try:
            os.close(file_descriptor)
        except OSError:
            # was already closed
            pass
        os.remove(tmp_path)


def _get_nb_to_py_mapping(
    root_dirs: Sequence[str], files: Optional[str], exclude: Optional[str]
) -> Dict[str, TemporaryFile]:
    """
    Get mapping between notebooks and temporary Python files.

    Parameters
    ----------
    root_dirs
        All the notebooks/directories passed in via the command-line.
    files
        Pattern of files to include.
    exclude
        Pattern of files to exclude.

    Returns
    -------
    Dict[str, Tuple[int, str]]
        Mapping between notebooks and temporary Python files.

    Raises
    ------
    FileNotFoundError
        If notebook isn't found.
    """
    nb_to_py_mapping: Dict[str, TemporaryFile] = {}
    for notebook in _get_all_notebooks(root_dirs, files, exclude):
        if not os.path.exists(notebook):
            _clean_up_tmp_files(nb_to_py_mapping)
            raise FileNotFoundError(
                f"{BOLD}No such file or directory: {notebook}{RESET}\n"
            )

        nb_to_py_mapping[notebook] = TemporaryFile(
            *tempfile.mkstemp(
                dir=os.path.dirname(notebook),
                prefix=remove_suffix(os.path.basename(notebook), ".ipynb"),
                suffix=".py",
            )
        )
        relative_path, _ = get_relative_and_absolute_paths(
            nb_to_py_mapping[notebook].file
        )
        nb_to_py_mapping[notebook] = nb_to_py_mapping[notebook]._replace(
            file=relative_path
        )
    return nb_to_py_mapping


def _main(  # pylint: disable=R0912,R0914,R0911
    cli_args: CLIArgs, configs: Configs
) -> int:
    """
    Run third-party tool on a single notebook or directory.

    Parameters
    ----------
    cli_args
        Commanline arguments passed to nbqa.
    configs
        Configuration passed to nbqa from commandline or via a config file

    Returns
    -------
    int
        Output code from third-party tool.
    """
    try:
        nb_to_py_mapping = _get_nb_to_py_mapping(
            cli_args.root_dirs, configs.nbqa_files, configs.nbqa_exclude
        )
    except FileNotFoundError as exc:
        sys.stderr.write(str(exc))
        return 1

    try:  # pylint disable=R0912

        if not nb_to_py_mapping:
            sys.stderr.write(
                "No .ipynb notebooks found in given directories: "
                f"{' '.join(i for i in cli_args.root_dirs if os.path.isdir(i))}\n"
            )
            return 0

        nb_info_mapping: MutableMapping[str, NotebookInfo] = {}

        for notebook, (file_descriptor, _) in nb_to_py_mapping.items():
            try:
                nb_info_mapping[notebook] = save_source.main(
                    notebook,
                    file_descriptor,
                    configs.nbqa_process_cells,
                    cli_args.command,
                    skip_bad_cells=configs.nbqa_skip_cells,
                )
            except Exception as exp:  # pylint: disable=W0703
                sys.stderr.write(
                    BASE_ERROR_MESSAGE.format(f"Error parsing {notebook}: {repr(exp)}")
                )
                return 1

        output, output_code, mutated = _run_command(
            cli_args.command,
            configs.nbqa_addopts,
            [i.file for i in nb_to_py_mapping.values()],
        )

        actually_mutated = False
        for notebook, (_, temp_python_file) in nb_to_py_mapping.items():
            output = _replace_temp_python_file_references_in_out_err(
                temp_python_file, notebook, output.out, output.err
            )
            output = map_python_line_to_nb_lines(
                cli_args.command,
                output.out,
                output.err,
                notebook,
                nb_info_mapping[notebook].cell_mappings,
            )

            if mutated:
                if not configs.nbqa_mutate and not configs.nbqa_diff:
                    # pylint: disable=C0301
                    msg = dedent(
                        f"""\
                        {BOLD}Mutation detected, will not reformat! Please use the `--nbqa-mutate` flag, e.g.:{RESET}

                            nbqa {cli_args.command} notebook.ipynb --nbqa-mutate

                        or, to only preview changes, use the `--nbqa-diff` flag, e.g.:

                            nbqa {cli_args.command} notebook.ipynb --nbqa-diff
                        """
                    )
                    # pylint: enable=C0301
                    sys.stderr.write(msg)
                    return 1

                try:
                    actually_mutated = (
                        REPLACE_FUNCTION[configs.nbqa_diff](
                            temp_python_file,
                            notebook,
                            nb_info_mapping[notebook],
                        )
                        or actually_mutated
                    )
                except Exception as exp:  # pylint: disable=W0703
                    sys.stderr.write(
                        BASE_ERROR_MESSAGE.format(
                            f"Error reconstructing {notebook}: {repr(exp)}"
                        )
                    )
                    return 1

        sys.stdout.write(output.out)
        sys.stderr.write(output.err)

        if mutated and not actually_mutated:
            output_code = 0
            mutated = False

        if configs.nbqa_diff:
            if mutated:
                sys.stdout.write(
                    "To apply these changes use `--nbqa-mutate` instead of `--nbqa-diff`\n"
                )
            return output_code

    finally:
        _clean_up_tmp_files(nb_to_py_mapping)
    return output_code


def _check_command_is_installed(command: str) -> None:
    """
    Check whether third-party tool is installed.

    Parameters
    ----------
    command
        Third-party tool being run on notebook(s).

    Raises
    ------
    ModuleNotFoundError
        If third-party tool isn't installed.
    UnsupportedPackageVersionError
        If third-party tool is of an unsupported version.
    """
    try:
        command_version = metadata.version(command)  # type: ignore
    except metadata.PackageNotFoundError:  # type: ignore
        try:
            import_module(command)
        except ImportError as exc:
            raise ModuleNotFoundError(_get_command_not_found_msg(command)) from exc
    else:
        if command in MIN_VERSIONS:
            min_version = MIN_VERSIONS[command]
            if parse_version(command_version) < parse_version(min_version):
                raise UnsupportedPackageVersionError(
                    command, command_version, min_version
                )


def main(argv: Optional[Sequence[str]] = None) -> int:
    """
    Run third-party tool (e.g. :code:`mypy`) against notebook or directory.

    Parameters
    ----------
    argv
        Command-line arguments (if calling this function directly), defaults to
        :code:`None` if calling via command-line.
    """
    cli_args: CLIArgs = CLIArgs.parse_args(argv)
    _check_command_is_installed(cli_args.command)
    project_root: Path = find_project_root(tuple(cli_args.root_dirs))
    configs: Configs = _get_configs(cli_args, project_root)
    configs.validate()

    return _main(cli_args, configs)


if __name__ == "__main__":
    sys.exit(main())



"""Import importlib.metadata, using backport if on an old version of Python."""

from importlib import import_module

try:
    # python 3.8 and above
    metadata = import_module("importlib.metadata")
except ImportError:  # pragma: nocover
    # (coverage is calculated using Python3.8, but we test for Python3.7 anyway)
    # python 3.7 and below
    metadata = import_module("importlib_metadata")



"""Parse output from code quality tools."""
import re
from functools import partial
from typing import Callable, Mapping, Match, NamedTuple, Sequence, Tuple, Union

from nbqa.path_utils import get_relative_and_absolute_paths


def _line_to_cell(match: Match[str], cell_mapping: Mapping[int, str]) -> str:
    """Replace Python line with corresponding Jupyter notebook cell."""
    return str(cell_mapping[int(match.group())])


class Output(NamedTuple):
    """Captured stdout and stderr."""

    out: str
    err: str


def _get_pattern(
    notebook: str, command: str, cell_mapping: Mapping[int, str]
) -> Sequence[Tuple[str, Union[str, Callable[[Match[str]], str]]]]:
    """
    Get pattern and substitutions with which to process code quality tool's output.

    Parameters
    ----------
    notebook
        Notebook command is being run on.
    command
        Code quality tool.
    cell_mapping
        Mapping from Python file lines to Jupyter notebook cells.

    Returns
    -------
    List
        Patterns and substitutions for reported output.
    """
    standard_substitution = partial(_line_to_cell, cell_mapping=cell_mapping)

    relative_path, absolute_path = get_relative_and_absolute_paths(notebook)

    if command == "black":
        return [
            (
                rf"(?<=^error: cannot format {re.escape(relative_path)}: Cannot parse: )\d+"
                rf"|(?<=^error: cannot format {re.escape(absolute_path)}: Cannot parse: )\d+",
                standard_substitution,
            ),
            (r"(?<=line )\d+(?=\)\nOh no! )", standard_substitution),
            (r"line cell_(?=\d+:\d+\)\nOh no! )", "cell_"),
        ]

    if command == "doctest":
        return [
            (
                rf'(?<=^File "{re.escape(relative_path)}", line )\d+'
                rf'|(?<=^File "{re.escape(absolute_path)}", line )\d+',
                standard_substitution,
            ),
            (
                rf'(?<=^File "{re.escape(relative_path)}",) line'
                rf'|(?<=^File "{re.escape(absolute_path)}",) line',
                "",
            ),
        ]

    # This is the most common one and is used by flake, pylint, mypy, and more.
    return [
        (
            rf"(?<=^{re.escape(absolute_path)}:)\d+"
            rf"|(?<=^{re.escape(relative_path)}:)\d+",
            standard_substitution,
        )
    ]


def map_python_line_to_nb_lines(
    command: str, out: str, err: str, notebook: str, cell_mapping: Mapping[int, str]
) -> Output:
    """
    Make sure stdout and stderr make reference to Jupyter Notebook cells and lines.

    Parameters
    ----------
    command
        Code quality tool.
    out
        Captured stdout from third-party tool.
    err
        Captured stdout from third-party tool.
    notebook
        Original Jupyter notebook.
    cell_mapping
        Mapping from Python file lines to Jupyter notebook cells.

    Returns
    -------
    Output
        Stdout, stderr with references to temporary Python file's lines replaced with references
        to notebook's cells and lines.
    """
    patterns = _get_pattern(notebook, command, cell_mapping)
    for pattern_, substitution_ in patterns:
        try:
            out = re.sub(pattern_, substitution_, out, flags=re.MULTILINE)
        except KeyError:
            pass
        try:
            err = re.sub(pattern_, substitution_, err, flags=re.MULTILINE)
        except KeyError:  # pragma: nocover (defensive check)
            pass

    return Output(out, err)



"""
Replace :code:`source` code cells of original notebook with ones from converted file.

The converted file will have had the third-party tool run against it by now.
"""

import copy
import json
import os
import sys
from difflib import unified_diff
from shutil import move
from typing import Any, Iterator, List, Mapping, MutableMapping, Sequence, Set

import tokenize_rt

from nbqa.handle_magics import NewMagicHandler
from nbqa.notebook_info import NotebookInfo
from nbqa.save_source import CODE_SEPARATOR
from nbqa.text import BOLD, GREEN, RED, RESET


def _restore_semicolon(
    source: str,
    cell_number: int,
    trailing_semicolons: Set[int],
) -> str:
    """
    Restore the semicolon at the end of the cell.

    Restore the trailing semicolon if the cell originally contained semicolon
    and the third party tool removed it.

    Parameters
    ----------
    source
        Portion of Python file between cell separators.
    cell_number
        Number of current cell.
    trailing_semicolons
        List of cells which originally had trailing semicolons.

    Returns
    -------
    str
        New source with removed semicolon restored.

    Raises
    ------
    AssertionError
        If code thought to be unreachable is reached.
    """
    if cell_number in trailing_semicolons:
        tokens = tokenize_rt.src_to_tokens(source)
        for idx, token in tokenize_rt.reversed_enumerate(tokens):
            if not token.src.strip(" \n") or token.name == "COMMENT":
                continue
            tokens[idx] = token._replace(src=token.src + ";")
            break
        else:  # pragma: nocover
            raise AssertionError(
                "Unreachable code, please report bug at https://github.com/nbQA-dev/nbQA/issues"
            )
        source = tokenize_rt.tokens_to_src(tokens)
    return source


def _reinstate_magics(
    source: str,
    temporary_lines: Sequence[NewMagicHandler],
) -> List[str]:
    """
    Put (preprocessed) line magics back in.

    Parameters
    ----------
    source
        Portion of Python file between cell separators.
    temporary_lines
        Mapping from temporary magic substitutions to original ipython magics

    Returns
    -------
    List[str]
        New source that can be saved into Jupyter Notebook.
    """
    for magic_substitution in temporary_lines:
        source = source.replace(magic_substitution.replacement, magic_substitution.src)
    return source.strip("\n").splitlines(True)


def _get_new_source(
    code_cell_number: int,
    notebook_info: NotebookInfo,
    pycell: str,
) -> List[str]:
    """
    Get new source to replace original one with.

    Parameters
    ----------
    code_cell_number
        Cell number (we start counting from 1).
    notebook_info
        Information about notebook cells used for processing
    pycell
        Cell from temporary Python script.

    Returns
    -------
    List
        New source for cell.
    """
    source = _restore_semicolon(
        pycell, code_cell_number, notebook_info.trailing_semicolons
    )
    return _reinstate_magics(
        source,
        notebook_info.temporary_lines.get(code_cell_number, []),
    )


def _get_pycells(python_file: str) -> Iterator[str]:
    """
    Parse cells from Python file.

    Parameters
    ----------
    python_file
        Temporary Python file notebook was converted to.

    Returns
    -------
    Iterator
        Parsed cells.
    """
    with open(python_file, encoding="utf-8") as handle:
        txt = handle.read()
    if txt.startswith(CODE_SEPARATOR):
        txt = txt[len(CODE_SEPARATOR) :]
    return iter(txt.split(CODE_SEPARATOR))


def _notebook_code_cells(
    notebook_json: Mapping[str, Any]
) -> Iterator[MutableMapping[str, Any]]:
    """
    Iterate through notebook's code cells.

    Parameters
    ----------
    notebook_json
        json-parsed notebook.

    Yields
    ------
    MutableMapping[str, Any]
        Cell content.
    """
    for cell in notebook_json["cells"]:
        if cell["cell_type"] == "code":
            yield cell


def mutate(python_file: str, notebook: str, notebook_info: NotebookInfo) -> bool:
    """
    Replace :code:`source` code cells of original notebook.

    Parameters
    ----------
    python_file
        Temporary Python file notebook was converted to.
    notebook
        Jupyter Notebook third-party tool is run against (unmodified).
    notebook_info
        Information about notebook cells used for processing

    Returns
    -------
    bool
        Whether mutation actually happened.
    """
    with open(notebook, encoding="utf-8") as handle:
        notebook_json = json.loads(handle.read())
    original_notebook_json = copy.deepcopy(notebook_json)

    pycells = _get_pycells(python_file)
    for code_cell_number, cell in enumerate(
        _notebook_code_cells(notebook_json), start=1
    ):
        if code_cell_number in notebook_info.code_cells_to_ignore:
            continue
        cell["source"] = _get_new_source(code_cell_number, notebook_info, next(pycells))

    if original_notebook_json == notebook_json:
        return False

    temp_notebook = os.path.join(
        os.path.dirname(python_file), os.path.basename(notebook)
    )
    with open(temp_notebook, "w", encoding="utf-8") as handle:
        handle.write(f"{json.dumps(notebook_json, indent=1, ensure_ascii=False)}\n")
    move(temp_notebook, notebook)
    return True


def _print_diff(code_cell_number: int, cell_diff: Iterator[str]) -> bool:
    """
    Print diff between cells, colouring as appropriate.

    Parameters
    ----------
    code_cell_number
        Current cell number
    cell_diff
        Diff between original and new versions of cell.

    Returns
    -------
    bool
        Whether non-null diff was printed.
    """
    line_changes = []
    for line in cell_diff:
        if line.startswith("+++") or line.startswith("---"):
            line_changes.append(line)
        elif line.startswith("+"):
            line_changes.append(f"{GREEN}{line}{RESET}")
        elif line.startswith("-"):
            line_changes.append(f"{RED}{line}{RESET}")
        else:
            line_changes.append(line)

    if line_changes:
        header = f"Cell {code_cell_number}"
        headers = [f"{BOLD}{header}{RESET}\n", f"{'-'*len(header)}\n"]
        sys.stdout.writelines(headers + line_changes + ["\n"])
        return True
    return False


def diff(python_file: str, notebook: str, notebook_info: NotebookInfo) -> bool:
    """
    View diff between new source of code cells and original sources.

    Parameters
    ----------
    python_file
        Temporary Python file notebook was converted to.
    notebook
        Jupyter Notebook third-party tool is run against (unmodified).
    notebook_info
        Information about notebook cells used for processing

    Returns
    -------
    bool
        Whether non-null diff was produced.
    """
    with open(notebook, encoding="utf-8") as handle:
        notebook_json = json.loads(handle.read())

    pycells = _get_pycells(python_file)

    actually_mutated = False

    for code_cell_number, cell in enumerate(
        _notebook_code_cells(notebook_json), start=1
    ):
        if code_cell_number in notebook_info.code_cells_to_ignore:
            continue
        new_source = _get_new_source(code_cell_number, notebook_info, next(pycells))
        cell["source"][-1] += "\n"
        new_source[-1] += "\n"
        cell_diff = unified_diff(
            cell["source"],
            new_source,
            fromfile=notebook,
            tofile=notebook,
        )
        actually_mutated = _print_diff(code_cell_number, cell_diff) or actually_mutated
    return actually_mutated
